import ast
from dataclasses import field, dataclass
from enum import Enum
from typing import Optional, Dict, List, Any

from transpiler.phases.typing.types import BaseType


class VarKind(Enum):
    """Kind of variable."""
    LOCAL = 1
    """`xxx = ...`"""
    GLOBAL = 2
    """`global xxx"""
    NONLOCAL = 3
    """`nonlocal xxx`"""
    SELF = 4
    OUTER_DECL = 5
    MODULE = 6


class VarType:
    pass


class RuntimeValue:
    pass


@dataclass
class VarDecl:
    kind: VarKind
    type: BaseType
    val: Any = RuntimeValue()


class ScopeKind(Enum):
    GLOBAL = 1
    """Global (module) scope"""
    FUNCTION = 2
    """Function scope"""
    FUNCTION_INNER = 3
    """Block (if, for, ...) scope inside a function"""
    CLASS = 4
    """Class scope"""


@dataclass
class Scope:
    parent: Optional["Scope"] = None
    kind: ScopeKind = ScopeKind.GLOBAL
    function: Optional["Scope"] = None
    global_scope: Optional["Scope"] = None
    vars: Dict[str, VarDecl] = field(default_factory=dict)
    children: List["Scope"] = field(default_factory=list)
    obj_type: Optional[BaseType] = None
    has_return: bool = False
    class_: Optional["Scope"] = None
    is_loop: Optional[ast.For | ast.While] = None

    @staticmethod
    def make_global():
        res = Scope()
        res.global_scope = res
        return res

    def is_in_loop(self) -> Optional[ast.For | ast.While]:
        if self.is_loop:
            return self.is_loop
        if self.parent is not None and self.kind != ScopeKind.FUNCTION:
            return self.parent.is_in_loop()
        return None

    def child(self, kind: ScopeKind):
        res = Scope(self, kind, self.function, self.global_scope)
        self.children.append(res)
        return res

    def declare_local(self, name: str, type: BaseType):
        """Declares a local variable"""
        self.vars[name] = VarDecl(VarKind.LOCAL, type)

    def get(self, name: str, kind: VarKind = VarKind.LOCAL) -> Optional[VarDecl]:
        """
        Gets the variable declaration of a variable in the current scope or any parent scope.
        """
        if (res := self.vars.get(name)) and res.kind == kind:
            return res
        if self.parent is not None:
            return self.parent.get(name, kind)
        return None
